// wios-check (Confirm sound device hardware capabilities.)
// compile with: g++ wios-check.cpp -lasound -o wios-check

#include <iostream>
#include <string>
#include <sstream>
#include <vector>

#ifdef __cplusplus
extern "C" {
#endif

#include <alsa/asoundlib.h>

#ifdef __cplusplus
}
#endif

class WiosSupports {
private:
  std::vector<snd_pcm_access_t> supported_accesses = {
    SND_PCM_ACCESS_RW_INTERLEAVED
  };

  std::vector<snd_pcm_format_t> supported_formats = {
    SND_PCM_FORMAT_U8,
    SND_PCM_FORMAT_S16_LE,
    SND_PCM_FORMAT_S24_LE,
    SND_PCM_FORMAT_S32_LE
  };

  std::string name;
  std::vector<snd_pcm_stream_t> stream;
  std::vector<snd_pcm_access_t> accesses;
  std::vector<snd_pcm_format_t> formats;
  std::string channels;
  std::string rates;

  std::stringstream ss;

public:
  WiosSupports(){}
  ~WiosSupports(){}
  bool set_stream(std::string dev, snd_pcm_stream_t& val){
    bool ret = false;
    if(val == SND_PCM_STREAM_PLAYBACK || val == SND_PCM_STREAM_CAPTURE){ ret = true; name = dev; stream.push_back(val); }
    return ret;
  }
  bool set_access(snd_pcm_access_t& val){
    bool ret = false;
    for(int i=0; i<supported_accesses.size(); i++){
      if(val == supported_accesses[i]){ ret = true; accesses.push_back(val); break; }
    }
    return ret;
  }
  bool set_format(snd_pcm_format_t& val){
    bool ret = false;
    for(int i=0; i<supported_formats.size(); i++){
      if(val == supported_formats[i]){ ret = true; formats.push_back(val); break; }
    }
    return ret;
  }
  bool set_channels(std::string& val){
    channels = val;
    return true;
  }
  bool set_rates(std::string& val){
    rates = val;
    return true;
  }
  bool generate_usage(){
    ss << "----------------------------------------\n";

    if(!stream.empty() && !accesses.empty() && !formats.empty() && !channels.empty() && !rates.empty()){

      ss << "You can use following device arguments with the wios command.\n";
      for(int i=0; i<stream.size(); i++){
        if(stream[i] == SND_PCM_STREAM_PLAYBACK){
          ss << "  In case of to use device " << name << " as a playback device.\n";
          ss << "    -p -x 0 -d " << name << "\n";
          ss << "      or\n";
          ss << "    -s -x 0 -y " << name << " -z <recording_device> (for the synchronous recording and playback)\n";
        }
        else if(stream[i] == SND_PCM_STREAM_CAPTURE){
          ss << "  In case of to use device " << name << " as a recording device.\n";
          ss << "    -r -x 0 -a " << name << "\n";
          ss << "      or\n";
          ss << "    -s -x 0 -y <playback_device> -z " << name << " (for the synchronous recording and playback)\n";
        }
      }
      ss << "\n";

      ss << "You can use following encoding argument with the wios command.\n";
      for(int i=0; i<formats.size(); i++){
        if(     formats[i] == SND_PCM_FORMAT_U8    ){ ss << "    -e u8 (same as --encoding u8)\n"; }
        else if(formats[i] == SND_PCM_FORMAT_S16_LE){ ss << "    -e 16 (same as --encoding 16)\n"; }
        else if(formats[i] == SND_PCM_FORMAT_S24_LE){ ss << "    -e 24 (same as --encoding 24)\n"; }
        else if(formats[i] == SND_PCM_FORMAT_S32_LE){ ss << "    -e 32 (same as --encoding 32)\n"; }
      }
      ss << "\n";

      ss << "You can use following channels argument with the wios command.\n";
      ss << "    -c <channels> (same as --channels <channels>)\n";
      ss << "    *) Please select one of " << channels << " as <channels>.\n\n";

      ss << "You can use following sampling rate argument with the wios command.\n";
      ss << "    -f <rate> (same as --frequency <rate>)\n";
      ss << "    *) Please select one of " << rates << " as <rate>.\n\n";

    }
    else{

      ss << "Currently the device " << name << " is not supported by wios. Although synchronous recording and playback by wios can not be used, there are the following alternatives. First, please record TSP instead of arecord or aplay. Next, create a transfer function using the feature to create a transfer function from the Mouth TSP of HARKTOOL5.\n\n";

    }
    std::cout << ss.str() << std::endl;
    return true;
  }
}; // class WiosSupports

class AudioDevice {
private:
  std::vector<snd_pcm_access_t> accesses = {
    SND_PCM_ACCESS_MMAP_INTERLEAVED,
    SND_PCM_ACCESS_MMAP_NONINTERLEAVED,
    SND_PCM_ACCESS_MMAP_COMPLEX,
    SND_PCM_ACCESS_RW_INTERLEAVED,
    SND_PCM_ACCESS_RW_NONINTERLEAVED
  };

  std::vector<snd_pcm_format_t> formats = {
    SND_PCM_FORMAT_S8,
    SND_PCM_FORMAT_U8,
    SND_PCM_FORMAT_S16_LE,
    SND_PCM_FORMAT_S16_BE,
    SND_PCM_FORMAT_U16_LE,
    SND_PCM_FORMAT_U16_BE,
    SND_PCM_FORMAT_S24_LE,
    SND_PCM_FORMAT_S24_BE,
    SND_PCM_FORMAT_U24_LE,
    SND_PCM_FORMAT_U24_BE,
    SND_PCM_FORMAT_S32_LE,
    SND_PCM_FORMAT_S32_BE,
    SND_PCM_FORMAT_U32_LE,
    SND_PCM_FORMAT_U32_BE,
    SND_PCM_FORMAT_FLOAT_LE,
    SND_PCM_FORMAT_FLOAT_BE,
    SND_PCM_FORMAT_FLOAT64_LE,
    SND_PCM_FORMAT_FLOAT64_BE,
    SND_PCM_FORMAT_IEC958_SUBFRAME_LE,
    SND_PCM_FORMAT_IEC958_SUBFRAME_BE,
    SND_PCM_FORMAT_MU_LAW,
    SND_PCM_FORMAT_A_LAW,
    SND_PCM_FORMAT_IMA_ADPCM,
    SND_PCM_FORMAT_MPEG,
    SND_PCM_FORMAT_GSM,
    SND_PCM_FORMAT_SPECIAL,
    SND_PCM_FORMAT_S24_3LE,
    SND_PCM_FORMAT_S24_3BE,
    SND_PCM_FORMAT_U24_3LE,
    SND_PCM_FORMAT_U24_3BE,
    SND_PCM_FORMAT_S20_3LE,
    SND_PCM_FORMAT_S20_3BE,
    SND_PCM_FORMAT_U20_3LE,
    SND_PCM_FORMAT_U20_3BE,
    SND_PCM_FORMAT_S18_3LE,
    SND_PCM_FORMAT_S18_3BE,
    SND_PCM_FORMAT_U18_3LE,
    SND_PCM_FORMAT_U18_3BE,
    SND_PCM_FORMAT_G723_24,
    SND_PCM_FORMAT_G723_24_1B,
    SND_PCM_FORMAT_G723_40,
    SND_PCM_FORMAT_G723_40_1B,
    SND_PCM_FORMAT_DSD_U8,
    SND_PCM_FORMAT_DSD_U16_LE,
    SND_PCM_FORMAT_DSD_U32_LE,
    SND_PCM_FORMAT_DSD_U16_BE,
    SND_PCM_FORMAT_DSD_U32_BE
  };

  std::vector<unsigned int> rates = {
    5512,
    8000,
    11025,
    16000,
    22050,
    32000,
    44100,
    48000,
    64000,
    88200,
    96000,
    176400,
    192000
  };

  std::string name;
  snd_pcm_t *pcm;
  snd_pcm_hw_params_t *hw_params;

  WiosSupports wios;
  std::stringstream wios_usage;

public:
  AudioDevice(): name(""), pcm(nullptr), hw_params(nullptr){
  }
  ~AudioDevice(){
    if(pcm != nullptr){
      snd_pcm_close(pcm);
      pcm = nullptr;
    }
    if(hw_params != nullptr){
      snd_pcm_hw_params_free(hw_params);
      hw_params = nullptr;
    }
  }

  bool is_valid(){
    if(pcm == nullptr){ return false; }
    if(name == ""){ return false; }
    if(hw_params == nullptr){ return false; }
    return true;
  }

  bool open_device(std::string& device_name, std::string& type, std::string& mode, bool reopen=true){
    int err = 0;
    snd_pcm_stream_t _type = SND_PCM_STREAM_PLAYBACK;
    int _mode = SND_PCM_NONBLOCK;
    if(type != "playback" && type != "capture"){
      std::cerr << "Cannot open device, because the device type is neither playback nor capture." << std::endl;
      return false;
    }
    if(mode != "non-block" && mode != "async"){
      std::cerr << "Cannot open device, because the device mode is neither non-block nor async." << std::endl;
      return false;
    }
    if(type == "capture"){ _type = SND_PCM_STREAM_CAPTURE; }
    if(mode == "async"){ _mode = SND_PCM_ASYNC; }
    if(pcm != nullptr){
      if(device_name == name){
        if(!reopen){
          std::cout << "Device '" << device_name << "' is already open, so it do nothing." << std::endl;
          return true;
        }
        std::cout << "Device '" << device_name << "' is already open, but 'reopen' has been specified, so it will close and reopen." << std::endl;
      }
      else{
        std::cout << "Device '" << device_name << "' different from the one previously opened was specified. Device '" << name << "' that was previously open are closed." << std::endl;
      }
      snd_pcm_close(pcm);
      name = device_name;
      err = snd_pcm_open(&pcm, device_name.c_str(), _type, _mode);
      wios.set_stream(device_name, _type);
    }
    else{
      name = device_name;
      err = snd_pcm_open(&pcm, device_name.c_str(), _type, _mode);
      wios.set_stream(device_name, _type);
    }
    if(err < 0){
      std::cerr << "Cannot open device '" << device_name << "': " << snd_strerror(err) << std::endl;
      name = "";
      pcm = nullptr;
      hw_params = nullptr;
      return false;
    }
    err = snd_pcm_hw_params_malloc(&hw_params);
    if(err < 0){
      std::cerr << "Cannot allocate hardware parameters: " << snd_strerror(err) << std::endl;
      snd_pcm_close(pcm);
      name = "";
      pcm = nullptr;
      hw_params = nullptr;
      return false;
    }
    err = snd_pcm_hw_params_any(pcm, hw_params);
    if(err < 0){
      std::cerr << "Cannot get hardware parameters: " << snd_strerror(err) << std::endl;
      snd_pcm_close(pcm);
      name = "";
      pcm = nullptr;
      snd_pcm_hw_params_free(hw_params);
      hw_params = nullptr;
      return false;
    }
    return true;
  }

  bool print_device(){
    if(!is_valid()){
      std::cerr << "Device is not ready." << std::endl;
      return false;
    }
    std::cout << "Device: " << name << " (type: " << snd_pcm_type_name(snd_pcm_type(pcm)) << ")" << std::endl;
    return true;
  }

  bool print_support_accesses(){
    std::stringstream ss;
    if(!is_valid()){
      std::cerr << "Device is not ready." << std::endl;
      return false;
    }
    ss << "Access types:";
    for(int i=0; i<accesses.size(); ++i){
      if(!snd_pcm_hw_params_test_access(pcm, hw_params, accesses[i])){
        ss << " " << snd_pcm_access_name(accesses[i]);
        wios.set_access(accesses[i]);
      }
    }
    std::cout << ss.str() << std::endl;
    return true;
  }

  bool print_support_formats(){
    std::stringstream ss;
    if(!is_valid()){
      std::cerr << "Device is not ready." << std::endl;
      return false;
    }
    ss << "Formats:";
    for(int i=0; i<formats.size(); ++i){
      if(!snd_pcm_hw_params_test_format(pcm, hw_params, formats[i])){
        ss << " " << snd_pcm_format_name(formats[i]);
        wios.set_format(formats[i]);
      }
    }
    std::cout << ss.str() << std::endl;
    return true;
  }

  bool print_support_channels(){
    int err = 0;
    unsigned int min = 0, max = 0;
    std::stringstream ss;
    if(!is_valid()){
      std::cerr << "Device is not ready." << std::endl;
      return false;
    }
    err = snd_pcm_hw_params_get_channels_min(hw_params, &min);
    if(err < 0){
      std::cerr << "Cannot get minimum channels count: " << snd_strerror(err) << std::endl;
      return false;
    }
    err = snd_pcm_hw_params_get_channels_max(hw_params, &max);
    if(err < 0){
      std::cerr << "Cannot get maximum channels count: " << snd_strerror(err) << std::endl;
      return false;
    }
    ss << "Channels:";
    if(min == max){
      ss << " " << min;
      std::string s("( " + std::to_string(min) + " )");
      wios.set_channels(s);
    }
    else{
      std::stringstream w; w << "(";
      bool seq = false;
      for(int i=min; i<=max; ++i){
        if(!snd_pcm_hw_params_test_channels(pcm, hw_params, i)){
          if(!seq){
            ss << " " << i; w << " " << i;
            seq = true;
          }
          else if(i == max){
            ss << "-" << i; w << "-" << i;
            seq = false;
          }
        }
        else{
          if(seq){
            ss << "-" << i-1; w << "-" << i-1;
            seq = false;
          }
        }
      }
      w << " )";
      std::string s(w.str());
      wios.set_channels(s);
    }
    std::cout << ss.str() << std::endl;
    return true;
  }

  bool print_support_rates(){
    int err = 0;
    unsigned int min = 0, max = 0;
    std::stringstream ss;
    if(!is_valid()){
      std::cerr << "Device is not ready." << std::endl;
      return false;
    }
    err = snd_pcm_hw_params_get_rate_min(hw_params, &min, NULL);
    if(err < 0){
      std::cerr << "Cannot get minimum rate: " << snd_strerror(err) << std::endl;
      return false;
    }
    err = snd_pcm_hw_params_get_rate_max(hw_params, &max, NULL);
    if(err < 0){
      std::cerr << "Cannot get maximum rate: " << snd_strerror(err) << std::endl;
      return false;
    }
    ss << "Sample rates:";
    if(min == max){
      ss << " " << min;
      std::string s("( " + std::to_string(min) + " )");
      wios.set_rates(s);
    }
    else if(!snd_pcm_hw_params_test_rate(pcm, hw_params, min + 1, 0)){
      ss << " " << min << "-" << max;
      std::string s("( " + std::to_string(min) + "-" + std::to_string(max) + " )");
      wios.set_rates(s);
    }
    else{
      std::stringstream w; w << "(";
      int any_rate = 0;
      for(int i=0; i<rates.size(); ++i){
        if(!snd_pcm_hw_params_test_rate(pcm, hw_params, rates[i], 0)){
          any_rate = 1;
          ss << " " << rates[i]; w << " " << rates[i];
        }
      }
      if(!any_rate){
        ss << " " << min << "-" << max; w << " " << min << "-" << max;
      }
      w << " )";
      std::string s(w.str());
      wios.set_rates(s);
    }
    std::cout << ss.str() << std::endl;
    return true;
  }

  bool print_interrupt_interval(){
    int err = 0;
    unsigned int min = 0, max = 0;
    std::stringstream ss;
    if(!is_valid()){
      std::cerr << "Device is not ready." << std::endl;
      return false;
    }
    err = snd_pcm_hw_params_get_period_time_min(hw_params, &min, NULL);
    if(err < 0){
      std::cerr << "Cannot get minimum period time: " << snd_strerror(err) << std::endl;
      return false;
    }
    err = snd_pcm_hw_params_get_period_time_max(hw_params, &max, NULL);
    if(err < 0){
      std::cerr << "Cannot get maximum period time: " << snd_strerror(err) << std::endl;
      return false;
    }
    ss << "Interrupt interval: " << min << "-" << max << " us";
    std::cout << ss.str() << std::endl;
    return true;
  }

  bool print_buffer_size(){
    int err = 0;
    unsigned int min = 0, max = 0;
    std::stringstream ss;
    if(!is_valid()){
      std::cerr << "Device is not ready." << std::endl;
      return false;
    }
    err = snd_pcm_hw_params_get_buffer_time_min(hw_params, &min, NULL);
    if(err < 0){
      std::cerr << "Cannot get minimum buffer time: " << snd_strerror(err) << std::endl;
      return false;
    }
    err = snd_pcm_hw_params_get_buffer_time_max(hw_params, &max, NULL);
    if(err < 0){
      std::cerr << "Cannot get maximum buffer time: " << snd_strerror(err) << std::endl;
      return false;
    }
    ss << "Buffer size: " << min << "-" << max << " us";
    std::cout << ss.str() << std::endl;
    return true;
  }

  bool print_usage(){
    wios.generate_usage();
    return true;
  }

}; // class AudioDevice

int main(int argc, char *argv[])
{
  std::string device_name("hw:0,0");
  std::string device_type("both");  // playback or capture or both
  std::string device_mode("non-block"); // non-block or async

  if(argc > 1){ device_name = std::string(argv[1]); }
  if(argc > 2){ device_type = std::string(argv[2]); }

  if(device_type == "playback" || device_type == "both"){
    std::string _device_type("playback");
    std::cout << "\nPlayback device\n==============================" << std::endl;
    AudioDevice dev;
    if(dev.open_device(device_name, _device_type, device_mode)){
      if(!dev.print_device()){ return 255; }
      if(!dev.print_support_accesses()){ return 255; }
      if(!dev.print_support_formats()){ return 255; }
      if(!dev.print_support_channels()){ return 255; }
      if(!dev.print_support_rates()){ return 255; }
      if(!dev.print_interrupt_interval()){ return 255; }
      if(!dev.print_buffer_size()){ return 255; }
    }
    dev.print_usage();
  }
  if(device_type == "capture" || device_type == "both"){
    std::string _device_type("capture");
    std::cout << "\nCapture device\n==============================" << std::endl;
    AudioDevice dev;
    if(dev.open_device(device_name, _device_type, device_mode)){
      if(!dev.print_device()){ return 255; }
      if(!dev.print_support_accesses()){ return 255; }
      if(!dev.print_support_formats()){ return 255; }
      if(!dev.print_support_channels()){ return 255; }
      if(!dev.print_support_rates()){ return 255; }
      if(!dev.print_interrupt_interval()){ return 255; }
      if(!dev.print_buffer_size()){ return 255; }
    }
    dev.print_usage();
  }

  return 0;
}

